from copy import deepcopy

import pytest
import requests_mock

from web3data.chains import Chains
from web3data.exceptions import APIError, EmptyResponseError
from web3data.handlers.base import BaseHandler

CHAINS = (
    Chains.BCH,
    Chains.BSV,
    Chains.BTC,
    Chains.ETH,
    Chains.ETH_RINKEBY,
    Chains.LTC,
    Chains.ZEC,
)
PARAMS = {"test": "value"}
HEADERS = {"foo": "bar"}
RESPONSE = {"baz": "qux"}


def assert_request_mock(m):
    assert m.call_count == 1
    assert (
        m.request_history[0].url == "http://example.com/test-route?test=value"
        or m.request_history[0].url
        == "http://example.com/test-route?test=value&format=csv"
    )
    # assert header kv pairs are in request headers
    assert set(HEADERS.items()).issubset(set(m.request_history[0].headers.items()))


@pytest.mark.parametrize("chain", CHAINS)
def test_base_handler_empty_response(chain):
    handler = BaseHandler(chain)
    with requests_mock.Mocker() as m:
        m.register_uri(requests_mock.ANY, requests_mock.ANY, text="")

        with pytest.raises(EmptyResponseError):
            handler.raw_query("http://example.com/", "test-route", HEADERS, PARAMS)

        assert m.call_count == 1
        assert_request_mock(m)


@pytest.mark.parametrize("chain", CHAINS)
def test_base_handler_empty_response_csv(chain):
    handler = BaseHandler(chain)
    with requests_mock.Mocker() as m:
        m.register_uri(requests_mock.ANY, requests_mock.ANY, text="")

        params = deepcopy(PARAMS)
        params["format"] = "csv"
        with pytest.raises(EmptyResponseError):
            handler.raw_query("http://example.com/", "test-route", HEADERS, params)

        assert m.call_count == 1
        assert_request_mock(m)


@pytest.mark.parametrize("chain", CHAINS)
def test_base_handler_empty_json_response(chain):
    handler = BaseHandler(chain)
    with requests_mock.Mocker() as m:
        m.register_uri(requests_mock.ANY, requests_mock.ANY, json={})

        with pytest.raises(EmptyResponseError):
            handler.raw_query("http://example.com/", "test-route", HEADERS, PARAMS)

        assert m.call_count == 1
        assert_request_mock(m)


@pytest.mark.parametrize("chain", CHAINS)
def test_base_handler_empty_csv_response(chain):
    handler = BaseHandler(chain)
    with requests_mock.Mocker() as m:
        m.register_uri(requests_mock.ANY, requests_mock.ANY, text="")

        params = deepcopy(PARAMS)
        params["format"] = "csv"
        with pytest.raises(EmptyResponseError):
            handler.raw_query("http://example.com/", "test-route", HEADERS, params)

        assert m.call_count == 1
        assert_request_mock(m)


@pytest.mark.parametrize("chain", CHAINS)
def test_base_handler_invalid_response(chain):
    handler = BaseHandler(chain)
    with requests_mock.Mocker() as m:
        m.register_uri(requests_mock.ANY, requests_mock.ANY, text="invalid")

        with pytest.raises(APIError):
            handler.raw_query("http://example.com/", "test-route", HEADERS, PARAMS)

        assert m.call_count == 1
        assert_request_mock(m)


@pytest.mark.parametrize("chain", CHAINS)
def test_base_handler_valid_response(chain):
    handler = BaseHandler(chain)
    with requests_mock.Mocker() as m:
        m.register_uri(requests_mock.ANY, requests_mock.ANY, json=RESPONSE)

        resp = handler.raw_query("http://example.com/", "test-route", HEADERS, PARAMS)

        assert resp == RESPONSE
        assert_request_mock(m)


@pytest.mark.parametrize("chain", CHAINS)
def test_base_handler_valid_response_csv(chain):
    handler = BaseHandler(chain)
    with requests_mock.Mocker() as m:
        m.register_uri(requests_mock.ANY, requests_mock.ANY, text="test")

        params = deepcopy(PARAMS)
        params["format"] = "csv"
        resp = handler.raw_query("http://example.com/", "test-route", HEADERS, params)

        assert resp == "test"
        assert_request_mock(m)



"""This package defines the sub-handlers and the main one."""

from web3data.handlers.address import AddressHandler
from web3data.handlers.api import APIHandler
from web3data.handlers.block import BlockHandler
from web3data.handlers.contract import ContractHandler
from web3data.handlers.market import MarketHandler
from web3data.handlers.signature import SignatureHandler
from web3data.handlers.token import TokenHandler
from web3data.handlers.transaction import TransactionHandler
from web3data.handlers.websocket import WebsocketHandler



"""Unit test package for web3data."""

from web3data.chains import Chains

API_PREFIX = "https://web3api.io/api/v2/"
HEADERS = {"foo": "bar", "baz": "qux"}
RESPONSE = {"test": "data"}
CHAINS = (
    Chains.BTC,
    Chains.BCH,
    Chains.BSV,
    Chains.ETH,
    Chains.ETH_RINKEBY,
    Chains.LTC,
    Chains.ZEC,
)



"""This module implements the websocket handler."""

import json
from typing import Iterable, Union
from uuid import uuid4

import websocket

from web3data.exceptions import APIError


class WebsocketHandler:
    """The subhandler for websocket-related queries."""

    def __init__(self, api_key: str, blockchain_id: str, url: str = None):
        """Return a new :code:`WebsocketHandler` instance.

        :param api_key: The API key to attach to payloads
        :param blockchain_id: The ID of the blockchain to query for
        :param url: The websocket server URL
        """

        self.api_key = api_key
        self.blockchain_id = blockchain_id
        self.url = url or "wss://ws.web3api.io/"

        self.expected_ids = set()  # internal IDs that still need confirmation
        self.internal_registry = {}  # internal ID -> payload and callback
        self.external_registry = {}  # subscription ID -> internal ID

        self.ws = websocket.WebSocketApp(
            self.url,
            on_message=lambda ws, message: self._on_message(ws, message),
            on_error=lambda ws, message: self._on_error(ws, message),
            on_close=lambda ws: self._on_close(ws),
            on_open=lambda ws: self._on_open(ws),
            header=[
                f"x-api-key: {self.api_key}",
                f"x-amberdata-blockchain-id: {self.blockchain_id}",
            ],
        )

    def _websocket_send(self, payload):
        """Send a message to the websocket server.

        :param payload: The payload to JSON serialize and send
        """
        self.ws.send(json.dumps(payload))  # pragma: no cover

    def register(self, params: Union[Iterable[str], str], callback=None):
        """Register a new event to listen for and its callback.

        This will subscribe to the given event identifiers and execute
        the provided callback function once the specified event is coming
        in. Please note that the listening and callback-handling routine
        only starts once the websocket client is running.

        The callback function should take two parameters:
        :code:`ws` and :code:`message`. The first parameter is the websocket
        client instance, which allows the user to update the client context.
        The latter argiment is the message, deserialized from the JSON object
        received by the websocket server.

        :param params: The event to subscribe to
        :param callback: The callback function to execute
        """
        params = (params,) if type(params) is str else params

        internal_id = str(uuid4())
        self.expected_ids.add(internal_id)
        self.internal_registry[internal_id] = {
            "callback": callback or (lambda: None),
            "payload": {
                "id": internal_id,
                "jsonrpc": "2.0",
                "method": "subscribe",
                "params": params,
            },
        }

    def unregister(self, external_id):
        """Unregister a subscription from the websocket server.

        Given an external ID (i.e. the subscription ID), this will
        remove the subscription data including locally stored payloads
        and identifiers. It will also trigger an unsubscribe message
        for the subscription being sent to the websocket server.

        :param external_id: The subscription ID to remove
        """
        internal_id = self.external_registry[external_id]
        del self.internal_registry[internal_id]
        del self.external_registry[external_id]

        internal_id = str(uuid4())
        self.expected_ids.add(internal_id)
        payload = {
            "jsonrpc": "2.0",
            "method": "unsubscribe",
            "params": [external_id],
            "id": internal_id,
        }
        self._websocket_send(payload)

    def run(self, **kwargs):
        """Run the websocket listening loop.

        This is a blocking endless loop that will send the subscription
        events to the websocket server, handle the responses, and then
        distribute the incoming messages across the registered callbacks.

        Any keyword arguments passed to this method are passed on to the
        websocket client's :code:`run_forever` method. Please consult the
        project's documentation for more details:
        https://pypi.org/project/websocket_client/

        :param kwargs: Additional arguments to pass to the websocket client
        """
        self.ws.run_forever(**kwargs)  # pragma: no cover

    def _on_message(self, ws, message):
        """An internal message handler to distribute responses.

        This internal callback will try to parse the incoming message and
        handle it based on whether it is a data message (in which case it will
        be routed to its respective subscription and its callback), a subscription
        acknowledgement message, or an unsubscription acknowledgement message.

        In the case of a subscription acknowledgement, the newly found subscription
        ID (aka external ID) is added to the client's mapping for future routing of
        data messages.

        In both latter cases, the given internal ID is removed from the internal
        set of messages to expect.

        :param ws: The websocket client instance
        :param message: The raw received message as serialized JSON
        """
        message = json.loads(message)

        if message.get("params"):
            # handle data message and execute user callback
            external_id = message.get("params", {}).get("subscription")
            internal_id = self.external_registry[external_id]
            subscription = self.internal_registry[internal_id]
            callback = subscription["callback"]
            callback(ws, message)
        elif type(message.get("result")) is str:
            # handle subscription acknowledgement
            internal_id = message.get("id")
            self.external_registry[message.get("result")] = internal_id
            self.expected_ids.remove(internal_id)
        elif type(message.get("result")) is bool:
            # handle unsubscription acknowledgement
            internal_id = message.get("id")
            self.expected_ids.remove(internal_id)
        else:
            raise APIError(f"Received unknown message: {message}")

    def _on_error(self, ws, error):
        """An internal handler for websocket errors.

        :param ws: The websocket client instance
        :param error: The error message
        """
        self.on_error(ws, error)

    def on_error(self, ws, error):
        """A user-defined handler for websocket errors.

        :param ws: The websocket client instance
        :param error: The error message
        """
        # to be overwritten by user if needed
        pass  # pragma: no cover

    def _on_close(self, ws):
        """An internal handler for websocket close events.

        :param ws: The websocket client instance
        """
        self.on_close(ws)

    def on_close(self, ws):
        """A user-defined handler for websocket close events.

        :param ws: The websocket client instance
        """
        # to be overwritten by user if needed
        pass  # pragma: no cover

    def _on_open(self, ws):
        """An internal handler for websocket open events.

        This handler will iterate over all internal identifiers
        and submit subscription a request for each. If no payload
        information can be found, a :code:`ValueError` is raised.

        After the requests have been sent, the user-defined on-open
        handler is called.

        :param ws: The websocket client instance
        """
        for internal_id in self.internal_registry.keys():
            payload = self.internal_registry.get(internal_id, {}).get("payload")
            if payload is None:
                raise ValueError(
                    f"Payload for internal ID {internal_id} does not exist"
                )
            self._websocket_send(payload)

        self.on_open(ws)

    def on_open(self, ws):
        """A user-defined handler for websocket open events.

        :param ws: The websocket client instance
        """
        # to be overwritten by user if needed
        pass  # pragma: no cover



"""This module contains the address subhandler."""

from typing import Dict

from web3data.chains import Chains
from web3data.handlers.base import BaseHandler


class TransactionHandler(BaseHandler):
    """The subhandler for transaction-related queries."""

    def __init__(self, initial_headers: Dict[str, str], chain: Chains):
        """Return a new :code:`TranscationHandler` instance.

        :param initial_headers: Base headers to attach to every request
        :param chain: The blockchain to fetch the information for
        """

        super().__init__(chain)
        self.initial_headers = initial_headers
        self.base_url = "https://web3api.io/api/v2/transactions/"

    def information(self, tx_hash: str, **kwargs) -> Dict:
        """Retrieves the transaction information for the specified hash.

        :param tx_hash: The transaction hash to fetch information for
        :key validationMethod: The validation method to be added to the response:
            none, basic, full. Default: none. (str)
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=tx_hash,
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_transfers(self, tx_hash: str) -> Dict:
        """Retrieves the token transfers that took place in the specified
        transaction.

        :param tx_hash: The transaction hash to fetch information for
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self.raw_query(
            base_url=self.base_url,
            route=f"{tx_hash}/token-transfers",
            headers=self.initial_headers,
            params={},
        )

    def find(self, **kwargs) -> Dict:
        """Retrieves all transactions matching the specified filters.

        :key status: Filter by the status of the transactions to retrieve
            (all, completed, failed, pending). (str)
        :key startDate: Filter by transactions executed after this date. Note that the interval
            can not exceed 1 minute (startDate and endDate should be both specified, or both empty) (int)
        :key endDate: Filter by transactions executed before this date. Note that the interval
            can not exceed 1 minute (startDate and endDate should be both specified, or both empty). (int)
        :key validationMethod: The validation method to be added to the response:
            none, basic, full. Default: none. (str)
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :key size: The number of results to return. (int)
        :key includeFunctions: Indicates whether or not to include log
            information for each transaction, if available (false|true) (bool)
        :key includeLogs: Indicates whether or not to include price information (false|true) (bool)
        :key includeTokenTransfers: Indicates whether or not to include token
            transfers information for each transaction, if available (false|true) (bool)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="",
            headers=self.initial_headers,
            params=kwargs,
        )

    def gas_percentiles(self, **kwargs) -> Dict:
        """Retrieves the latest gas price percentiles for the transactions.

        :key numBlocks: Number of past blocks on which to base the percentiles. (int)
        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self.raw_query(
            base_url=self.base_url,
            route="gas/percentiles",
            headers=self.initial_headers,
            params=kwargs,
        )

    def gas_predictions(self) -> Dict:
        """Retrieves the latest gas predictions for the transactions.

        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self.raw_query(
            base_url=self.base_url,
            route="gas/percentiles",
            headers=self.initial_headers,
            params={},
        )

    def metrics(self) -> Dict:
        """Get metrics for recent confirmed transactions for a given
        blockchain.

        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="metrics/latest",
            headers=self.initial_headers,
            params={},
        )

    def volume(self, **kwargs) -> Dict:
        """Retrieves the historical (time series) volume of transactions.

        :param kwargs: Additional query parameter options
        :key timeFormat: The time format to use for the timestamps: milliseconds/ms or iso/iso861. (str)
        :key timeFrame: The time frame to return the historical data in:
            by day (1d, 2d, ..., all),
            by hour (1h, 2h, ..., 72h) or
            by minute (1m, 2m, ..., 360m) (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="volume",
            headers=self.initial_headers,
            params=kwargs,
        )



import pytest
import requests_mock

from web3data.chains import Chains
from web3data.exceptions import APIError
from web3data.handlers import APIHandler
from web3data.handlers.address import AddressHandler
from web3data.handlers.block import BlockHandler
from web3data.handlers.contract import ContractHandler
from web3data.handlers.market import MarketHandler
from web3data.handlers.signature import SignatureHandler
from web3data.handlers.token import TokenHandler
from web3data.handlers.transaction import TransactionHandler

TEST_KEY = "test-key"
TEST_ID = "test-id"
TEST_RPC = {"test": "response"}
ALLOWED_CHAINS = (Chains.BTC, Chains.ETH, Chains.ETH_RINKEBY)
NON_RPC_CHAINS = (
    Chains.BCH,
    Chains.BSV,
    Chains.LTC,
    Chains.ZEC,
)


@pytest.mark.parametrize(
    "handler",
    (
        APIHandler(TEST_KEY, TEST_ID, chain=Chains.BCH),
        APIHandler(TEST_KEY, TEST_ID, chain=Chains.BSV),
        APIHandler(TEST_KEY, TEST_ID, chain=Chains.BTC),
        APIHandler(TEST_KEY, TEST_ID, chain=Chains.ETH),
        APIHandler(TEST_KEY, TEST_ID, chain=Chains.ETH_RINKEBY),
        APIHandler(TEST_KEY, TEST_ID, chain=Chains.LTC),
        APIHandler(TEST_KEY, TEST_ID, chain=Chains.ZEC),
    ),
)
def test_api_handler_initialized(handler):
    assert isinstance(handler.address, AddressHandler)
    assert isinstance(handler.block, BlockHandler)
    assert isinstance(handler.contract, ContractHandler)
    assert isinstance(handler.market, MarketHandler)
    assert isinstance(handler.signature, SignatureHandler)
    assert isinstance(handler.token, TokenHandler)
    assert isinstance(handler.transaction, TransactionHandler)


@pytest.mark.parametrize("chain", ALLOWED_CHAINS)
def test_allowed_rpc(chain):
    handler = APIHandler(TEST_KEY, TEST_ID, chain)
    with requests_mock.Mocker() as m:
        m.register_uri(requests_mock.ANY, requests_mock.ANY, json=TEST_RPC)

        response = handler.rpc("test-method", ["test-param"])
        assert m.call_count == 1
        assert response == TEST_RPC
        assert m.request_history[0].url == "https://rpc.web3api.io/?x-api-key=test-key"
        assert m.request_history[0].method == "POST"
        assert m.request_history[0].json() == {
            "id": 1,
            "jsonrpc": "2.0",
            "method": "test-method",
            "params": ["test-param"],
        }


@pytest.mark.parametrize("chain", NON_RPC_CHAINS)
def test_non_rpc(chain):
    handler = APIHandler(TEST_KEY, TEST_ID, chain)
    with pytest.raises(APIError):
        handler.rpc("test-method", ["test-param"])



"""This module contains the main API handler class."""

from typing import Any, List

import requests

from web3data import __version__
from web3data.chains import Chains
from web3data.exceptions import APIError
from web3data.handlers.address import AddressHandler
from web3data.handlers.block import BlockHandler
from web3data.handlers.contract import ContractHandler
from web3data.handlers.market import MarketHandler
from web3data.handlers.signature import SignatureHandler
from web3data.handlers.token import TokenHandler
from web3data.handlers.transaction import TransactionHandler
from web3data.handlers.websocket import WebsocketHandler


class APIHandler:
    """The API handler object for client requests."""

    def __init__(self, api_key: str, blockchain_id: str, chain: Chains):
        """Return a new API handler instance.

        :param api_key: The API key to attach to request headers
        :param blockchain_id: The ID of the blockchain to query for
        :param chain: The enum value for the blockchain to query for
        """

        self.api_key = api_key
        self.blockchain_id = blockchain_id
        self.chain = chain

        # TODO: Validation
        headers = {
            "x-api-key": self.api_key,
            "x-amberdata-blockchain-id": self.blockchain_id,
            "User-Agent": f"web3data-py v{__version__}",
        }
        self.address = AddressHandler(headers, chain)
        self.token = TokenHandler(headers, chain)
        self.contract = ContractHandler(headers, chain)
        self.transaction = TransactionHandler(headers, chain)
        self.block = BlockHandler(headers, chain)
        self.signature = SignatureHandler(headers, chain)
        self.market = MarketHandler(headers, chain)
        self.websocket = WebsocketHandler(
            api_key=self.api_key, blockchain_id=self.blockchain_id
        )

    def rpc(self, method: str, params: List[str], ident: int = 1):
        """Perform an HTTP POST RPC call on the API.

        Consult the docs here for further details on supported commands:
        https://docs.amberdata.io/reference#rpc-overview

        :param method: The RPC method to call
        :param params: Parameters attached to the RPC call
        :param ident: RPC call identifier
        """
        if self.chain not in (Chains.ETH, Chains.ETH_RINKEBY, Chains.BTC):
            raise APIError(f"RPC calls are not supported for {self.chain}")

        return requests.post(
            "https://rpc.web3api.io/",
            json={
                "jsonrpc": "2.0",
                "method": method,
                "params": params,
                "id": ident,
            },
            headers={"x-amberdata-blockchain-id": self.blockchain_id},
            params={"x-api-key": self.api_key},
        ).json()



"""This module contains the address subhandler."""

from typing import Dict

from web3data.chains import Chains
from web3data.handlers.base import BaseHandler


class ContractHandler(BaseHandler):
    """The subhandler for contract-related queries."""

    def __init__(self, initial_headers: Dict[str, str], chain: Chains):
        """Return a new :code:`ContractHandler` instance.

        :param initial_headers: Base headers to attach to every request
        :param chain: The blockchain to fetch the information for
        """

        super().__init__(chain)
        self.initial_headers = initial_headers
        self.base_url = "https://web3api.io/api/v2/contracts/{hash}/"

    def _contract_query(
        self,
        route: str,
        headers: Dict[str, str],
        params: Dict[str, str],
        address: str = "",
    ) -> Dict:
        """Helper method for contract-related API queries.

        :param route: The endpoint route to query
        :param headers: The headers to attach to the request
        :param params: The request's query parameters
        :param address: The contract address string to query for
        :return: The parsed API response
        """

        self._check_chain_supported()
        return self.raw_query(
            base_url=self.base_url.format(hash=address),
            route=route,
            headers=headers,
            params=params,
        )

    def audit(self, address: str, **kwargs) -> Dict:
        """Retrieves the vulnerabilities audit for the specified contract (if
        available).

        The automated security checks are provided by MythX. Check out their
        stellar service over at https://mythx.io/.

        :param address: The address to fetch information for
        :param kwargs: Additional query parameter options
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._contract_query(
            address=address,
            route="audit",
            headers=self.initial_headers,
            params=kwargs,
        )

    def details(self, address: str, **kwargs) -> Dict:
        """Retrieves all the detailed information for the specified contract
        (ABI, bytecode, sourcecode...).

        :param address: The address to fetch information for
        :param kwargs: Additional query parameter options
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._contract_query(
            address=address,
            route="",
            headers=self.initial_headers,
            params=kwargs,
        )

    def functions(self, address: str, **kwargs) -> Dict:
        """Retrieves the functions of the specified contract (if available).

        If not available on chain, the byte code is decompiled and a list
        of functions is extracted from it.

        :param address: The address to fetch information for
        :param kwargs: Additional query parameter options
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._contract_query(
            address=address,
            route="functions",
            headers=self.initial_headers,
            params=kwargs,
        )



"""This module contains the address subhandler."""

from typing import Any, Dict

from web3data.chains import Chains
from web3data.handlers.base import BaseHandler


class BlockHandler(BaseHandler):
    """The subhandler for block-related queries."""

    def __init__(self, initial_headers: Dict[str, str], chain: Chains):
        """Return a new :code:`BlockHandler` instance.

        :param initial_headers: Base headers to attach to every request
        :param chain: The blockchain to fetch the information for
        """

        super().__init__(chain)
        self.initial_headers = initial_headers
        self.base_url = "https://web3api.io/api/v2/blocks/{id}/"

    def _block_query(
        self,
        route: str,
        headers: Dict[str, str],
        params: Dict[str, Any],
        block_id: str = "",
    ) -> Dict:
        """Helper method for block-related API queries.

        :param route: The endpoint route to query
        :param headers: The headers to attach to the request
        :param params: The request's query parameters
        :param block_id: The block ID to query for
        :return: The parsed API response
        """
        return self.raw_query(
            base_url=self.base_url.format(id=block_id),
            route=route,
            headers=headers,
            params=params,
        )

    def single(self, block_id: str, **kwargs) -> Dict:
        """Retrieves the block specified by its id (number or hash).

        :param block_id: The block's ID to fetch information for
        :key validationMethod: The validation method to be added to the response: none, basic, full.
            Default: none. (str)
        :key timeFormat: The time format to use for the timestamps (milliseconds, ms, iso, iso8611). (str)
        :return: The API response parsed into a dict
        """
        return self._block_query(
            block_id=block_id,
            route="",
            headers=self.initial_headers,
            params=kwargs,
        )

    def total(self, **kwargs) -> Dict:
        """Retrieves all the blocks within the specified range.

        :key startNumber: The range of blocks to return, inclusive (startNumber and endNumber
            should be both specified, or both empty) (str)
        :key endNumber: The end of the range of blocks to return, exclusive (startNumber and endNumber
            should be both specified, or both empty) (str)
        :key size: The number of results to return. (int)
        :key validationMethod: The validation method to be added to the response:
            none, basic, full. Default: none. (str)
        :return: The API response parsed into a dict
        """
        return self._block_query(route="", headers=self.initial_headers, params=kwargs)

    def functions(self, block_id: str, **kwargs) -> Dict:
        """Retrieves all the functions which were called at the specified block
        number or hash.

        :param block_id: The block's ID to fetch information for
        :key validationMethod: The validation method to be added to the response:
            none, basic, full. Default: none. (str)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._block_query(
            block_id=block_id,
            route="functions",
            headers=self.initial_headers,
            params=kwargs,
        )

    def logs(self, block_id: str, **kwargs) -> Dict:
        """Retrieves all the logs at the specified block number or hash.

        :param block_id: The block's ID to fetch information for
        :key validationMethod: The validation method to be added to the response:
            none, basic, full. Default: none. (str)
        :key transactionHash: Filter by logs for this transaction. (str)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._block_query(
            block_id=block_id,
            route="logs",
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_transfers(self, block_id: str, **kwargs) -> Dict:
        """Retrieves all the token which were transferred at the specified
        block number.

        :param block_id: The block's ID to fetch information for
        :key amount: Filter by tokens transfers where the number of tokens is equal
            to the specified amount. (int)
        :key amountGt: Filter by token transfers where the number of tokens is more than
            the specified amount. (int)
        :key amountGte: Filter by token transfers where the number of tokens is more
            than or equal to the specified amount. (int)
        :key amountLt: Filter by token transfers where the number of tokens is less
            than the specified amount. (int)
        :key amountLte: Filter by token transfers where the number of tokens is less
            than or equal to the specified amount. (int)
        :key from: Filter by token transfers originating from this address. (str)
        :key to: Filter token transfers received by this address. (str)
        :key tokenAddress: Filter by token transfers for this token. (str)
        :key transactionHash: Filter by token transfers for this transaction. (str)
        :key includePrice: Indicates whether or not to include price data with the results. (bool)
        :key currency: The currency of the price information. Options: usd, btc.
            Only used in conjunction with includePrice. (str)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._block_query(
            block_id=block_id,
            route="token-transfers",
            headers=self.initial_headers,
            params=kwargs,
        )

    def transactions(self, block_id: str, **kwargs) -> Dict:
        """Retrieves all the transactions included in a specified block id.

        :param block_id: The block's ID to fetch information for
        :key includeFunctions: Indicates whether or not to include functions (aka internal messages)
            information for each transaction, if available. Options: true, false. (bool)
        :key includeTokenTransfers: Indicates whether or not to include token transfers information
            for each transaction, if available. Options: true, false. (bool)
        :key includeLogs: Indicates whether or not to include log information for each transaction,
            if available. Options: true, false. (bool)
        :key validationMethod: The validation method to be added to the response: none, basic, full.
            Default: none. (str)
        :key currency: The currency of the price information. Options: usd, btc. Only used in
            conjunction with includePrice. (str)
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key startDate: Filter by transactions executed after this date. Note that the interval
            can not exceed 1 minute (startDate and endDate should be both specified, or both empty) (int)
        :key endDate: Filter by transactions executed before this date. Note that the interval
            can not exceed 1 minute (startDate and endDate should be both specified, or both empty). (int)
        :key size: The number of results to return. (int)
        :return: The API response parsed into a dict
        """
        return self._block_query(
            block_id=block_id,
            route="transactions",
            headers=self.initial_headers,
            params=kwargs,
        )

    def metrics_latest(self, **kwargs) -> Dict:
        """Get metrics for recent confirmed blocks for a given blockchain.

        :key timeFormat: The time format to use for the timestamps (milliseconds, ms, iso, iso8611). (str)
        :return: The API response parsed into a dict
        """
        return self._block_query(
            route="metrics/latest", headers=self.initial_headers, params=kwargs
        )

    def metrics_historical(self, **kwargs) -> Dict:
        """Get metrics for historical confirmed blocks for a given blockchain.

        :key timeFormat: The time format to use for the timestamps (milliseconds, ms, iso, iso8611). (str)
        :key timeFrame: time frame to return the historical data in, options: (1m, 5m, 1h, 1d, 1w) (str)
        :key startDate: Filter by data after this date. (str)
        :key endDate: Filter by data before this date. (str)
        :return: The API response parsed into a dict
        """
        return self._block_query(
            route="metrics/historical",
            headers=self.initial_headers,
            params=kwargs,
        )



"""This module contains the main API client class."""

from web3data.chains import Chains
from web3data.handlers.api import APIHandler


class Web3Data:
    """The Amberdata API client object."""

    def __init__(self, api_key: str):
        """Return a new API client instance.

        :param api_key: The Amberdata API key to perform requests with
        """
        self.btc = APIHandler(
            api_key=api_key,
            blockchain_id="408fa195a34b533de9ad9889f076045e",
            chain=Chains.BTC,
        )
        self.bch = APIHandler(
            api_key=api_key,
            blockchain_id="43b45e71cc0615b491cb699e7071fc06",
            chain=Chains.BCH,
        )
        self.bsv = APIHandler(
            api_key=api_key,
            blockchain_id="a818635d36dbe125e26167c4438e2217",
            chain=Chains.BSV,
        )
        self.eth = APIHandler(
            api_key=api_key, blockchain_id="1c9c969065fcd1cf", chain=Chains.ETH
        )
        self.eth_rinkeby = APIHandler(
            api_key=api_key,
            blockchain_id="1b3f7a72b3e99c13",
            chain=Chains.ETH_RINKEBY,
        )
        self.ltc = APIHandler(
            api_key=api_key,
            blockchain_id="f94be61fd9f4fa684f992ddfd4e92272",
            chain=Chains.LTC,
        )
        self.zec = APIHandler(
            api_key=api_key,
            blockchain_id="b7d4f994f33c709be4ce6cbae31d7b8e",
            chain=Chains.ZEC,
        )



"""This module contains an enum with names of supported chains."""

from enum import Enum


class Chains(Enum):
    """Blockchains supported by the Amberdata API."""

    BTC = 1
    BCH = 2
    BSV = 3
    ETH = 4
    ETH_RINKEBY = 5
    LTC = 6
    ZEC = 7



import pytest

from web3data.client import Web3Data
from web3data.handlers import APIHandler

CLIENT = Web3Data("test-key")


@pytest.mark.parametrize(
    "client",
    (
        CLIENT.bch,
        CLIENT.bsv,
        CLIENT.btc,
        CLIENT.eth,
        CLIENT.eth_rinkeby,
        CLIENT.ltc,
        CLIENT.zec,
    ),
)
def test_client_initialized(client):
    assert isinstance(client, APIHandler)
    assert isinstance(client, APIHandler)
    assert isinstance(client, APIHandler)
    assert isinstance(client, APIHandler)
    assert isinstance(client, APIHandler)
    assert isinstance(client, APIHandler)
    assert isinstance(client, APIHandler)
    assert isinstance(client, APIHandler)
    assert isinstance(client, APIHandler)



from itertools import product

import pytest
import requests_mock

from web3data.chains import Chains
from web3data.exceptions import APIError
from web3data.handlers.contract import ContractHandler

from . import API_PREFIX, CHAINS, HEADERS, RESPONSE

LIMITED_CHAINS = (
    Chains.BCH,
    Chains.BSV,
    Chains.BTC,
    Chains.LTC,
    Chains.ZEC,
)
CONTRACT_HANDLER_METHODS = (
    ["audit", ("ADDRESS",), LIMITED_CHAINS],
    ["details", ("ADDRESS",), LIMITED_CHAINS],
    ["functions", ("ADDRESS",), LIMITED_CHAINS],
)

CONTRACT_PARAMS = []
for chain_value, call in product(CHAINS, CONTRACT_HANDLER_METHODS):
    CONTRACT_PARAMS.append([chain_value] + call[:2] + [chain_value in call[2]])


@pytest.mark.parametrize("chain,method,parameters,raises", CONTRACT_PARAMS)
def test_contract_handler(chain, method, parameters, raises):
    handler = ContractHandler(initial_headers=HEADERS, chain=chain)
    method = getattr(handler, method, None)

    if raises:
        with pytest.raises(APIError):
            method(*parameters)
    else:
        with requests_mock.Mocker() as m:
            m.register_uri(requests_mock.ANY, requests_mock.ANY, json=RESPONSE)
            response = method(*parameters)

            assert m.call_count == 1
            assert response == RESPONSE
            assert m.request_history[0].url.startswith(API_PREFIX)
            # assert header kv pairs are in request headers
            assert set(HEADERS.items()).issubset(
                set(m.request_history[0].headers.items())
            )



"""This module contains the address subhandler."""

from typing import Any, Dict, List

from web3data.chains import Chains
from web3data.handlers.base import BaseHandler


class AddressHandler(BaseHandler):
    """The subhandler for address-related queries."""

    def __init__(self, initial_headers: Dict[str, str], chain: Chains):
        """Return a new :code:`AddressHandler` instance.

        :param initial_headers: Base headers to attach to every request
        :param chain: The blockchain to fetch the information for
        """

        super().__init__(chain=chain)
        self.initial_headers = initial_headers
        self.chain = chain
        self.base_url = "https://web3api.io/api/v2/addresses/{hash}/"

    def _address_query(
        self,
        route: str,
        headers: Dict[str, str],
        params: Dict[str, Any],
        address: str = "",
    ) -> Dict:
        """Base method to perform address queries.

        :param route: The endpoint route to query
        :param headers: The headers to attach to the request
        :param params: The request's query parameters
        :param address: The address string to query for
        :return: The parsed API response
        """
        return self.raw_query(
            base_url=self.base_url.format(hash=address),
            route=route,
            headers=headers,
            params=params,
        )

    def total(self, **kwargs) -> Dict:
        """Retrieves every Ethereum address that has been seen on the network.

        :key hash: Filter by a specific address (str)
        :key blockNumber: Filter by addresses first encountered at this block number (int)
        :key blockNumberGt: Filter by addresses first encountered after this block number,
            not including this blocknumber (int)
        :key blockNumberGte: Filter by addresses first encountered after this block number,
            including this block number (str)
        :key blockNumberLt: Filter by addresses first encountered before this block number,
            not including this block number (int)
        :key blockNumberLte: Filter by addresses first encountered before this block number,
            including this block number (str)
        :key startDate: Filter by addresses first encountered after this date (int)
        :key endDate: Filter by addresses first encountered after before date (int)
        :key type: Filter by addresses of the specified type (EOA or CONTRACT) (str)
        :key transactionHash: Filter by addresses first encountered at this transaction hash (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page. (int)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            route="", headers=self.initial_headers, params=kwargs
        )

    def adoption(self, address: str, **kwargs) -> Dict:
        """Retrieves the historical adoption for the specified address.

        :param address: The address to fetch information for
        :key timeFormat: The time format to use with the timestamps: milliseconds/ms or iso/iso8611 (str)
        :key timeFrame: The time frame to return the historical data in (str):
            by day (1d, 2d, ..., all),
            by hour (1h, 2h, ..., 72h) or
            by minute (1m, 2m, ..., 360m)
        :key timePeriod: The time period (in minutes) to aggregate the historical data (str)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="adoption",
            headers=self.initial_headers,
            params=kwargs,
        )

    def balance_historical(self, address: str, **kwargs) -> Dict:
        """Retrieves the historical (time series) account balances for the
        specified address.

        :param address: The address to fetch information for
        :key blockNumber: Filter by account balances at block number (int)
        :key startDate: Filter by account balances which happened after this date (int)
        :key endDate: Filter by account balances which happened before this date (int)
        :key value: Filter by account balances where the balance is equal to this value (int)
        :key valueGt: Filter by account balances where the balance is greater than this value (int)
        :key valueGte: Filter by account balances where the balance is greater than or equal to this value (int)
        :key valueLt: Filter by account balances where the balance is less than this value (int)
        :key valueLte: Filter by account balances where the balance is less than or equal to this value (int)
        :key includePrice: Indicates whether or not to include price data with the results. Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page (int)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="account-balances/historical",
            headers=self.initial_headers,
            params=kwargs,
        )

    def balance_latest(self, address: str, **kwargs) -> Dict:
        """Retrieves the current account balance for the specified address.

        :param address: The address to fetch information for
        :key includePrice: Indicates whether or not to include price data with the results. Options: true, false. (str)
        :key currency: The currency of the price information. Options: usd, btc.
            Only used in conjunction with includePrice. (str)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="account-balances/latest",
            headers=self.initial_headers,
            params=kwargs,
        )

    def balances_batch(self, addresses: List[str], **kwargs) -> Dict:
        """Retrieves the latest account and token balances for the specified
        addresses.

        This is super useful if you want to get an entire portfolio's summary in a single call.
        Get totals for ETH & all token amounts with market prices.

        :param addresses: The addresses to fetch information for
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :key timeFormat: The time format to use for the timestamps (milliseconds, ms, iso, iso8611). (str)
        :return: The API response parsed into a dict
        """
        kwargs.update({"addresses": ",".join(addresses)})
        return self._address_query(
            route="balances", headers=self.initial_headers, params=kwargs
        )

    def balances(self, address: str, **kwargs) -> Dict:
        """Retrieves the latest account and token balances for the specified
        address.

        :param address: The address to fetch information for
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :key timeFormat: The time format to use for the timestamps (milliseconds, ms, iso, iso8611). (str)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="balances",
            headers=self.initial_headers,
            params=kwargs,
        )

    def information(self, address: str, **kwargs) -> Dict:
        """Retrieves information about the specified address.

        This includes network(s) and blockchain(s) this address exist within.

        :param address: The address to fetch information for
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="information",
            headers=self.initial_headers,
            params=kwargs,
        )

    def internal_messages(self, address: str, **kwargs) -> Dict:
        """Retrieves internal messages where this address is either the
        originator or a recipient.

        :param address: The address to fetch information for
        :key blockNumber: Filter by internal messages contained within this block number (int)
        :key from: Filter by internal messages for this "from" address (str)
        :key to: Filter by internal messages for this "to" address (str)
        :key transactionHash: Filter by internal messages for this transaction (str)
        :key startDate: Filter by internal messages which happened after this date (int)
        :key endDate: Filter by internal messages which happened before this date (int)
        :key validationMethod: The validation method to be added to the response: none, basic, full.
            Default: none. (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page (int)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._address_query(
            address=address,
            route="functions",
            headers=self.initial_headers,
            params=kwargs,
        )

    def logs(self, address: str, **kwargs) -> Dict:
        """Retrieves the logs for the transactions where this address is either
        the originator or a recipient.

        :param address: The address to fetch information for
        :key blockNumber: Filter by logs contained in this block number (int)
        :key startDate: Filter by logs which happened after this date (int)
        :key endDate: Filter by logs which happened before this date (int)
        :key topic: Filter by logs containing this topic (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page (int)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._address_query(
            address=address,
            route="logs",
            headers=self.initial_headers,
            params=kwargs,
        )

    def metadata(self, address: str, **kwargs) -> Dict:
        """Retrieves statistics about the specified address: balances,
        holdings, etc.

        :param address: The address to fetch information for
        :key timeFormat: The time format to use for the timestamps (milliseconds, ms, iso, iso8611). (str)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="metadata",
            headers=self.initial_headers,
            params=kwargs,
        )

    def pending_transactions(self, address: str, **kwargs) -> Dict:
        """Retrieves pending transactions the specified address is involved in.

        :param address: The address to fetch information for
        :key from: Filter by transactions for this "from" address. (str)
        :key to: Filter by transactions for this "to" address (str)
        :key startDate: Filter by transactions which happened after this date. (int)
        :key endDate: Filter by transactions which happened before this date. (int)
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :key page: The page number to return. (int)
        :key size: The number of records per page. (int)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="pending-transactions",
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_balances_historical(self, address: str, **kwargs) -> Dict:
        """Retrieves the historical (time series) token balances for the
        specified address.

        :param address: The address to fetch information for
        :key amount: Filters by token balances which value is equal to this amount (int)
        :key amountGt: Filter by token balances which value is greater than this amount (int)
        :key amountGte: Filter by token balances which value is greater than or equal to this amount (int)
        :key amountLt: Filter by token balances which value is less than this amount (int)
        :key amountLte: Filter by token balances which value is less than or equal to this amount (int)
        :key tokenHolder: Filter by token balances which are held by this address (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page (int)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._address_query(
            address=address,
            route="token-balances",
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_balances_latest(self, address: str, **kwargs) -> Dict:
        """Retrieves the tokens this address is holding.

        :param address: The address to fetch information for
        :key direction: The direction by which to sort the tokens (ascending or descending). (str)
        :key includePrice: Indicates whether or not to include price data with the results. Options: true, false. (str)
        :key currency: The currency of the price information (usd or eth.)
            - only used in conjunction with includePrice. (str)
        :key sortType: The metric by which to rank the tokens (amount, name, symbol). (str)
        :key page: The page number to return. (str)
        :key size: The number of records per page. (str)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._address_query(
            address=address,
            route="tokens",
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_transfers(self, address: str, **kwargs) -> Dict:
        """Retrieves all token transfers involving the specified address.

        :param address: The address to fetch information for
        :key amount: Filter by token transfers which value is equal to this amount. (int)
        :key amountGt: Filter by token transfers which value is greater than this amount. (int)
        :key amountGte: Filter by token transfers which value is greater than or equal to this amount. (int)
        :key amountLt: Filter by token transfers which value is less than this amount. (int)
        :key amountLte: Filter by token transfers which value is less than or equal to this amount (int)
        :key blockNumber: Filter by token transfers with this block number. (int)
        :key recipientAddress: Filter by token transfers which recipient is the specified address. (str)
        :key senderAddress: Filter by token transfers which sender is the specified address. (str)
        :key startDate: Filter by token transfers which happened after this date. (int)
        :key endDate: Filter by token transfers which happened before this date. (int)
        :key tokenAddress: Filter by token transfers for this token. (str)
        :key transactionHash: Filter by token transfers for this transaction hash. (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page. (int)
        :key validationMethod: The validation method to be added to the response: none, basic, full.
            Default: none. (str)
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information. Options: usd, btc.
            Only used in conjunction with includePrice. (str)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self._address_query(
            address=address,
            route="token-transfers",
            headers=self.initial_headers,
            params=kwargs,
        )

    def transactions(self, address: str, **kwargs) -> Dict:
        """Retrieves the transactions where this address was either the
        originator or a recipient.

        :param address: The address to fetch information for
        :key blockNumber: Filter by transactions for this block number. (int)
        :key from: Filter by transactions for this "from" address. (str)
        :key to: Filter by transactions for this "to" address (str)
        :key startDate: Filter by transactions which happened after this date. (date)
        :key endDate: Filter by transactions which happened before this date. (date)
        :key validationMethod: The validation method to be added to the response: none, basic, full.
            Default: none. (str)
        :key includeFunctions: Indicates whether or not to include functions (aka internal messages)
            information for each transaction, if available (false|true). (bool)
        :key includeLogs: Indicates whether or not to include log information for each transaction,
            if available (false|true). (bool)
        :key includeTokenTransfers: Indicates whether or not to include token transfers information for
            each transaction, if available (false|true). (bool)
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :key page: The page number to return. (int)
        :key size: The number of records per page. (int)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="transactions",
            headers=self.initial_headers,
            params=kwargs,
        )

    def usage(self, address: str, **kwargs) -> Dict:
        """Retrieves the historical usage for the specified address.

        :param address: The address to fetch information for
        :key timeFormat: The time format to use with the timestamps: milliseconds/ms or iso/iso8611 (str)
        :key timeFrame: The time frame to return the historical data in:
            by day (1d, 2d, ..., all),
            by hour (1h, 2h, ..., 72h) or
            by minute (1m, 2m, ..., 360m) (str)
        :key timePeriod: The time period (in minutes) to aggregate the historical data. (str)
        :return: The API response parsed into a dict
        """
        return self._address_query(
            address=address,
            route="usage",
            headers=self.initial_headers,
            params=kwargs,
        )

    def metrics(self) -> Dict:
        """Get metrics for all addresses that have exist publicly for a given
        blockchain.

        Default metrics are for Ethereum over a 24h period.

        :return: The API response parsed into a dict
        """
        return self._address_query(
            route="metrics/latest", headers=self.initial_headers, params={}
        )



"""This module contains the address subhandler."""

from typing import Dict

from web3data.chains import Chains
from web3data.handlers.base import BaseHandler


class TokenHandler(BaseHandler):
    """The subhandler for token-related queries."""

    def __init__(self, initial_headers: Dict[str, str], chain: Chains):
        """Return a new :code:`TokenHandler` instance.

        :param initial_headers: Base headers to attach to every request
        :param chain: The blockchain to fetch the information for
        """

        super().__init__(chain)
        self.initial_headers = initial_headers

    def _token_query(self, address: str, route: str, params: Dict[str, str]):
        """Helper method for token-related API queries.

        :param route: The endpoint route to query
        :param params: The request's query parameters
        :param address: The token address string to query for
        :return: The parsed API response
        """
        self._check_chain_supported()
        return self.raw_query(
            base_url=f"https://web3api.io/api/v2/tokens/{address}/",
            route=route,
            headers=self.initial_headers,
            params=params,
        )

    def holders_historical(self, address: str, **kwargs) -> Dict:
        """Retrieves the historical (time series) token holders for the
        specified token address.

        :param address: The token's smart contract address
        :key holderAddresses: A comma separated list of addresses for which the historical
            holdings are to be retrieved. (str)
        :key timeFormat: The time format to use for the timestamps: milliseconds/ms or iso/iso861. (str)
        :key timeFrame: The time frame to return the historical data in:
            by day (1d, 2d, ..., all),
            by hour (1h, 2h, ..., 72h),
            by minute (1m, 2m, ..., 360m) or
            by tick (1t, 10t, ..., 1000t). (str)
        :key includePrice: Indicates whether or not to include price data
            with the results. (bool)
        :key currency: The currency of the price information. Options: usd, btc.
            Only used in conjunction with includePrice. (str)
        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self._token_query(address, "holders/historical", kwargs)

    def holders_latest(self, address: str, **kwargs) -> Dict:
        """Retrieves the token holders for the specified address.

        :param address: The token's smart contract address
        :key numTokens: Filter by token holders who own the specified amount
            of tokens. (int)
        :key numTokensGt: Filter by token holders who own more than the
            specified amount of tokens (int)
        :key numTokensGte: Filter by token holders who own more than or equal
            to the specified amount of tokens (int)
        :key numTokensLt: Fitler by token holders who own less than the
            specified amount of tokens (int)
        :key numTokensLte: Filter by token holders who own less than or equal
            to the specified amount of tokens (int)
        :key tokenAddress: Filter by token holders for this token (mandatory) (str)
        :key timestampGt: Filter by token holders who started holding the token
            after the specified date (int)
        :key timestampGte: Filter by token holders who started holding the token
            after or equal to the specified date (int)
        :key timestampLt: Filter by token holders who started holding the token
            before the specified date (int)
        :key timestampLte: Filter by token holders who started holding the token
            before or equal to the specified date (int)
        :key includePrice: Indicates whether or not to include price data with
            the results. Options: true, false. (bool)
        :key currency: The currency of the price information (usd or btc.)
            - only used in conjunction with includePrice. (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page (int)
        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self._token_query(address, "holders/latest", kwargs)

    def supply_historical(self, address: str, **kwargs) -> Dict:
        """Retrieves the historical token supplies (and derivatives) for the
        specified address.

        :param address: The token's smart contract address
        :param kwargs: Additional query parameter options
        :key timeFormat: The time format to use for the timestamps:
            milliseconds/ms or iso/iso861. (str)
        :key timeInterval: The time interval to return the historical data in:
            by day (days) or by hour (hours). (str)
        :key startDate: Filter by token prices after this date
            - note that the interval can not exceed 6 months (d), or 30 days (h). (int)
        :key endDate: Filter by token prices before this date
            - note that the interval can not exceed 6 months (d), or 30 days (h). (int)
        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self._token_query(address, "supplies/historical", kwargs)

    def supply_latest(self, address: str) -> Dict:
        """Retrieves the latest token supplies (and derivatives) for the
        specified address.

        :param address: The token's smart contract address
        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self._token_query(address, "supplies/latest", {})

    def transfers(self, address: str, **kwargs) -> Dict:
        """Retrieves all token transfers involving the specified address.

        :param address: The token's smart contract address
        :key amount: Filter by token transfers which value is equal to this amount. (int)
        :key amountGt: Filter by token transfers which value is greater than this amount. (int)
        :key amountGte: Filter by token transfers which value is greater than or equal
            to this amount. (int)
        :key amountLt: Filter by token transfers which value is less than this amount. (int)
        :key amountLte: Filter by token transfers which value is less than or equal
            to this amount (int)
        :key blockNumber: Filter by token transfers with this block number. (int)
        :key recipientAddress: Filter by token transfers which recipient is the
            specified address. (str)
        :key senderAddress: Filter by token transfers which sender is the
            specified address. (str)
        :key startDate: Filter by token transfers which happened after this date. (int)
        :key endDate: Filter by token transfers which happened before this date. (int)
        :key tokenAddress: Filter by token transfers for this token. (str)
        :key transactionHash: Filter by token transfers for this transaction hash. (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page. (int)
        :key validationMethod: The validation method to be added to the response:
            none, basic, full. Default: none. (str)
        :key includePrice: Indicates whether or not to include price data with the results.
            Options: true, false. (bool)
        :key currency: The currency of the price information. Options: usd, btc.
            Only used in conjunction with includePrice. (str)
        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self._token_query(address, "transfers", kwargs)

    def velocity(self, address: str, **kwargs) -> Dict:
        """Retrieves the historical velocity for the specified address.

        :param address: The token's smart contract address
        :key timeFormat: The time format to use with the timestamps: milliseconds/ms or iso/iso8611 (str)
        :key timeFrame: The time frame to return the historical data in:
            by day (1d, 2d, ..., all),
            by hour (1h, 2h, ..., 72h) or
            by minute (1m, 2m, ..., 360m) (str)
        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self._token_query(address, "velocity", kwargs)

    def volume(self, address: str, **kwargs) -> Dict:
        """Retrieves the historical number of transfers for the specified
        address.

        :param address: The token's smart contract address
        :key timeFormat: The time format to use with the timestamps: milliseconds/ms or iso/iso8611 (str)
        :key timeFrame: The time frame to return the historical data in:
            by day (1d, 2d, ..., all),
            by hour (1h, 2h, ..., 72h) or
            by minute (1m, 2m, ..., 360m) (str)
        :return: The API response parsed into a dict
        """
        self._check_chain_supported()
        return self._token_query(address, "volume", kwargs)



import json
from json.decoder import JSONDecodeError
from types import FunctionType
from unittest.mock import Mock

import pytest

from web3data.exceptions import APIError
from web3data.handlers.websocket import WebsocketHandler

DATA_RESPONSE = json.dumps(
    {
        "jsonrpc": "2.0",
        "method": "subscription",
        "params": {
            "result": {
                "blockchainId": "1c9c969065fcd1cf",
                "number": 7452758,
                "hash": "0x087ad0362e03cc781b2f252333d68caa52575bc72976d6350d7608561c57770a",
            },
            "subscription": "e0b0f42177341bff16987e1b12904d7dc8a6e4417dee0291fa134b5044763482",
        },
    }
)

SUBSCRIPTION_RESPONSE = json.dumps(
    {
        "jsonrpc": "2.0",
        "id": 1,
        "result": "242d29d5c0ec9268f51a39aba4ed6a36c757c03c183633568edb0531658a9799",
    }
)
UNSUBSCRIPTION_RESPONSE = json.dumps({"jsonrpc": "2.0", "id": 1, "result": True})
UNKNOWN_RESPONSE = json.dumps(
    {"jsonrpc": "2.0", "id": 1, "result": {"invalid": "datatype"}}
)


def get_handler():
    handler = WebsocketHandler("test-key", "test-id")
    handler._websocket_send = Mock()
    return handler


def assert_handler_initialized(handler: WebsocketHandler):
    assert handler.url == "wss://ws.web3api.io/"
    assert not handler.expected_ids
    assert handler.external_registry == {}
    assert handler.internal_registry == {}


def test_register():
    handler = get_handler()
    assert_handler_initialized(handler)

    handler.register("test", lambda: None)
    internal_entry = list(handler.internal_registry.values())[0]

    assert handler.expected_ids
    assert handler.external_registry == {}
    assert handler.internal_registry != {}
    assert type(internal_entry["callback"]) == FunctionType
    assert internal_entry["payload"]["jsonrpc"] == "2.0"
    assert internal_entry["payload"]["method"] == "subscribe"
    assert internal_entry["payload"]["params"] == ("test",)


def test_unregister():
    handler = get_handler()
    assert_handler_initialized(handler)

    handler.internal_registry["internal-id"] = {"important": "things"}
    handler.external_registry["external-id"] = "internal-id"
    handler.unregister("external-id")

    assert not handler.internal_registry
    assert not handler.external_registry
    assert len(handler.expected_ids) == 1
    handler._websocket_send.assert_called_once()


def test_on_message_data():
    handler = get_handler()
    callback_mock = Mock()
    assert_handler_initialized(handler)

    # mock the registration callback
    handler.register("test", callback_mock)
    # connect external to internal ID
    handler.external_registry[
        "e0b0f42177341bff16987e1b12904d7dc8a6e4417dee0291fa134b5044763482"
    ] = list(handler.internal_registry.keys())[0]

    handler._on_message(None, DATA_RESPONSE)

    callback_mock.assert_called_with(None, json.loads(DATA_RESPONSE))


def test_on_message_subscription():
    handler = get_handler()
    assert_handler_initialized(handler)
    # add request ID to expected responses
    handler.expected_ids = {1}

    handler._on_message(None, SUBSCRIPTION_RESPONSE)

    # acknowledges, we don't expect another internal ID to come in
    assert handler.expected_ids == set()
    # successfully linked external to internal ID
    assert (
        handler.external_registry[
            "242d29d5c0ec9268f51a39aba4ed6a36c757c03c183633568edb0531658a9799"
        ]
        == 1
    )


def test_on_message_unsubscription():
    handler = get_handler()
    assert_handler_initialized(handler)
    # add request ID to expected responses
    handler.expected_ids = {1}

    handler._on_message(None, UNSUBSCRIPTION_RESPONSE)

    # acknowledges, we don't expect another internal ID to come in
    assert handler.expected_ids == set()


def test_on_message_unknown():
    handler = get_handler()
    assert_handler_initialized(handler)

    with pytest.raises(APIError):
        handler._on_message(None, UNKNOWN_RESPONSE)


def test_on_message_invalid():
    handler = get_handler()
    assert_handler_initialized(handler)

    with pytest.raises(JSONDecodeError):
        handler._on_message(None, "invalid")


@pytest.mark.parametrize(
    "hook,args",
    (("on_open", (None,)), ("on_close", (None,)), ("on_error", (None, None))),
)
def test_custom_callbacks(hook, args):
    handler = get_handler()
    setattr(handler, hook, Mock())

    # simulate callbacks from websocket client
    getattr(handler, f"_{hook}")(*args)

    getattr(handler, hook).assert_called_with(*args)


def test_subscriptions_on_open():
    handler = get_handler()
    mock_callback = Mock()
    assert_handler_initialized(handler)

    handler.register(("test", "foo"), mock_callback)

    handler._on_open(None)

    handler._websocket_send.assert_called_once()


def test_error_on_open():
    handler = get_handler()
    assert_handler_initialized(handler)

    # subscription without payload
    handler.internal_registry[1] = {}

    with pytest.raises(ValueError):
        handler._on_open(None)



"""This module contains API-related exceptions."""


class APIError(Exception):
    """An exception denoting generic API errors.

    This error is raised when the API returns invalid response data, like invalid JSON.
    """

    pass


class EmptyResponseError(APIError):
    """An exception denoting an empty API response.

    This error is raised when the API response content is empty, or it contains an empty JSON
    object.
    """

    pass



"""This module contains the API handler's base class."""

from json.decoder import JSONDecodeError
from typing import Dict, Union

import requests
from requests.compat import urljoin

from web3data.chains import Chains
from web3data.exceptions import APIError, EmptyResponseError


class BaseHandler:
    """The API handler base class.

    This class defines the basic methods of performing REST API endpoint queries as well as RPC
    queries, which are implemented across all handler classes to standardize API requests.
    """

    LIMITED = (
        Chains.BTC,
        Chains.BCH,
        Chains.BSV,
        Chains.LTC,
        Chains.ZEC,
    )

    def __init__(self, chain: Chains):
        self.chain = chain

    def _check_chain_supported(self):
        if self.chain in self.LIMITED:
            raise APIError(f"This method is not supported for {self.chain}")

    @staticmethod
    def raw_query(
        base_url: str,
        route: str,
        headers: Dict[str, str],
        params: Dict[str, str],
    ) -> Union[Dict, str]:
        """Perform an HTTP GET request on an API REST endpoint.

        :param base_url: The API base URL (common prefix)
        :param route: The endpoint route after the base (variable suffix)
        :param headers: Headers to attach to the API request
        :param params: Query parameters to attach to the URL
        :return: The API response parsed into a dict
        """
        resp = requests.get(
            url=urljoin(base_url, route), headers=headers, params=params
        )

        if not resp.content:
            # triggered if the API returns empty response body
            raise EmptyResponseError("The API returned an empty JSON response")

        if params.get("format", "") == "csv":
            return resp.text

        try:
            result = resp.json()
        except JSONDecodeError:
            # triggered e.g. when API returns empty response or XML error message
            raise APIError(f"Unable to parse API response to JSON: {resp.content}")
        if not result:
            # triggered if the API returns empty JSON object response
            raise EmptyResponseError("The API returned an empty JSON response")

        return result



"""This module contains the address subhandler."""

from typing import Dict

from web3data.chains import Chains
from web3data.handlers.base import BaseHandler


class SignatureHandler(BaseHandler):
    """The subhandler for signature-related queries."""

    def __init__(self, initial_headers: Dict[str, str], chain: Chains):
        """Return a new :code:`SignatureHandler` instance.

        :param initial_headers: Base headers to attach to every request
        :param chain: The blockchain to fetch the information for
        """

        super().__init__(chain)
        self.initial_headers = initial_headers
        self.base_url = "https://web3api.io/api/v2/signatures/"

    def details(self, signature: str) -> Dict:
        """Retrieves detailed information about the specified signature hash.

        :param signature: The signature string to look up
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self.raw_query(
            base_url=self.base_url,
            route=signature,
            headers=self.initial_headers,
            params={},
        )



"""This module contains the address subhandler."""

import warnings
from typing import Dict

from web3data.chains import Chains
from web3data.handlers.base import BaseHandler


class MarketHandler(BaseHandler):
    """The subhandler for market-related queries."""

    def __init__(self, initial_headers: Dict[str, str], chain: Chains):
        """Return a new :code:`MarketHandler` instance.

        :param initial_headers: Base headers to attach to every request
        :param chain: The blockchain to fetch the information for
        """

        super().__init__(chain)
        self.initial_headers = initial_headers
        self.base_url = "https://web3api.io/api/v2/"

    def exchanges(self, **kwargs) -> Dict:
        """Retrieves information about supported exchange-pairs.

        These types of data are supported:
        - ticker
        - ohlc (open-high-low-close)
        - trade
        - order_book
        - order_book_update

        :key exchange: only return data for the given exchanges (comma separated) (str)
        :key pair: only return data for the given pairs (comma separated) (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="market/exchanges",
            headers=self.initial_headers,
            params=kwargs,
        )

    def ohlcv(self, **kwargs) -> Dict:
        """Retrieves information about supported exchange-pairs for ohlcv.

        :key exchange: Filter by data for the given exchanges (comma separated). (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="market/ohlcv/information",
            headers=self.initial_headers,
            params=kwargs,
        )

    def pairs(self, **kwargs) -> Dict:
        """Retrieves information about supported exchange-pairs. These types of
        data are supported:

        - ticker
        - ohlc (open-high-low-close)
        - trade
        - order_book
        - order_book_update

        :key exchange: only return data for the given exchanges (comma separated) (str)
        :key pair: only return data for the given pairs (comma separated) (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="market/pairs",
            headers=self.initial_headers,
            params=kwargs,
        )

    def rankings(self, **kwargs) -> Dict:
        """Retrieves the top ranked assets by a specific metric.

        :key direction: The sort order in which assets are ranked ascending
            or descending. (str)
        :key sortType: The metric used to rank the assets.
            Options: changeInPrice, currentPrice, liquidMarketCap, marketCap,
            tokenVelocity, tradeVolume, transactionVolume, uniqueAddresses (str)
        :key timeInterval: The time interval in which to return
            the historical data days or hours. (str)
        :key type: The type(s) of assets to include in the rankings:
            erc20|, erc721, erc777, erc884, erc998. Note: leaving this
            parameter empty means all tokens will be included. (str)
        :key page: The page number to return. (int)
        :key size: The number of records per page. (int)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="market/rankings",
            headers=self.initial_headers,
            params=kwargs,
        )

    def price_pairs(self) -> Dict:
        """Retrieves the assets for which latest prices are available.

        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="market/prices/pairs",
            headers=self.initial_headers,
            params={},
        )

    def ticker_pairs(self, **kwargs) -> Dict:
        """Retrieves the list of all available market tickers.

        :key exchange: Only return data for the given exchanges (comma separated). (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="market/tickers/information",
            headers=self.initial_headers,
            params=kwargs,
        )

    def trades(self, **kwargs) -> Dict:
        """Retrieves the list of all available market trade data sets.

        :key exchange: Only return data for the given exchanges (comma separated). (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route="market/trades/information",
            headers=self.initial_headers,
            params=kwargs,
        )

    def order_best_bid_historical(self, pair: str, **kwargs) -> Dict:
        """Retrieves historical best bid and offer information for the
        specified pair.

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve order book data.
            Example: exchange=bitfinex,bitstamp (str)
        :key timeFormat: The timestamp format to use for the timestamps:
            milliseconds/ms or iso/iso8611. (str)
        :key startDate: Filter by pairs after this date. (int)
        :key endDate: Filter by pairs before this date. (int)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/orders/{pair}/bbo/historical",
            headers=self.initial_headers,
            params=kwargs,
        )

    def order_best_bid_latest(self, pair: str, **kwargs) -> Dict:
        """Retrieves the latest best bid and offer information for the
        specified pair and exchange.

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve order book data.
            Example: exchange=bitfinex,bitstamp (str)
        :key timeFormat: The timestamp format to use for the timestamps:
            milliseconds/ms or iso/iso8611. (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/orders/{pair}/bbo",
            headers=self.initial_headers,
            params=kwargs,
        )

    def order_book_updates(self, pair: str, **kwargs) -> Dict:
        """

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve order book data.
            Example: exchange=bitfinex,bitstamp (str)
        :key timeFormat: The timestamp format to use for the timestamps:
            milliseconds/ms or iso/iso8611. (str)
        :key startDate: Filter by pairs after this date. (int)
        :key endDate: Filter by pairs before this date. (int)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/orders/{pair}/update",
            headers=self.initial_headers,
            params=kwargs,
        )

    def order_book(self, pair: str, **kwargs) -> Dict:
        """Retrieves the order book data for the specified pair.

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve order book data.
            Example: exchange=bitfinex,bitstamp (str)
        :key timestamp: The timestamp at which to return the order book information
            (closest match, lower or equal to the timestamp specified). (str)
        :key timeFormat: The timestamp format to use for the timestamps:
            milliseconds/ms or iso/iso8611. (str)
        :key startDate: Filter by pairs after this date. Formats:
            milliseconds, iso, or iso8611 (str)
        :key endDate: Filter by pairs before this date. Formats:
            milliseconds, iso, or iso8611.
            Note: Must be greater than startDate and cannot exceed 10 minutes. (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/orders/{pair}",
            headers=self.initial_headers,
            params=kwargs,
        )

    def uniswap_liquidity(self, pair: str, **kwargs) -> Dict:
        """Retrieves the Uniswap-specific ether and token balance pairs over
        time.

        Note: This endpoint returns a max of 6 months of historical data. In order
        to get more than 6 months you must use the startDate & endDate parameters
        to move the time frame window to get the next n days/months of data.

        :param pair: The asset pair to look up
        :key timeFormat: The timestamp format to use for the timestamps:
            milliseconds/ms or iso/iso8611. (str)
        :key startDate: Filter to liquidity changes after this date (int)
        :key endDate: Filter to liquidity changes before this date (int)
        :return: The API response parsed into a dict
        """
        warnings.warn(
            "The Uniswap liquidity endpoint is scheduled for deprecation",
            PendingDeprecationWarning,
        )
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/orders/uniswap/{pair}/liquidity",
            headers=self.initial_headers,
            params=kwargs,
        )

    def trade_pairs_historical(self, pair: str, **kwargs) -> Dict:
        """Retrieves the historical (time series) trade data for the specified
        pair.

        Note: This endpoint returns a max of 6 months of historical data. In order to
        get more than 6 months you must use the startDate & endDate parameters to move
        the time frame window to get the next n days/months of data.

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve market trades. (str)
        :key startDate: Filter by trades after this date. (int)
        :key endDate: Filter by trades before this date. (int)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/trades/{pair}/historical",
            headers=self.initial_headers,
            params=kwargs,
        )

    def ohlcv_pair_latest(self, pair: str, **kwargs) -> Dict:
        """Retrieves the latest open-high-low-close for the specified pair.

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve OHLCV.
            Example: exchange=bitfinex,bitstamp (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/ohlcv/{pair}/latest",
            headers=self.initial_headers,
            params=kwargs,
        )

    def ohlcv_pair_historical(self, pair: str, **kwargs) -> Dict:
        """Retrieves the historical (time series) open-high-low-close for the
        specified pair.

        Note: This endpoint returns a max of 6 months of historical data. In order to get more
        than 6 months you must use the startDate & endDate parameters to move the time frame
        window to get the next n days/months of data.

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve OHLCV.
            Example: exchange=bitfinex,bitstamp (str)
        :key startDate: Filter by pairs after this date. (int)
        :key endDate: Filter by pairs before this date. (int)
        :key timeInterval: Time interval to return the historical data
            in ("days" | "hours" | "minutes") (str)
        :key timeFormat: Time format to use for the timestamps
            ( "milliseconds" | "ms" | "iso" | "iso8611" ) (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/ohlcv/{pair}/historical",
            headers=self.initial_headers,
            params=kwargs,
        )

    def price_pair_historical(self, pair: str, **kwargs) -> Dict:
        """Retrieves the historical prices for the specified asset.

        :param pair: The asset pair to look up
        :key timeFormat: Time format to use for the timestamps
            (Options: milliseconds, ms, iso, iso8611) (str)
        :key startDate: Filter by prices after this date. (str)
        :key endDate: Filter by prices before this date. (str)
        :key timeInterval: Time interval to return the historical data
            in ("days" | "hours" | "minutes") (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/spot/prices/pairs/{pair}/historical",
            headers=self.initial_headers,
            params=kwargs,
        )

    def price_pair_latest(self, pair: str, **kwargs) -> Dict:
        """Retrieves the latest price for the specified asset.

        :param pair: The asset pair to look up
        :key timeFormat: Time format to use for the timestamps
            (Options: milliseconds, ms, iso, iso8611) (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/spot/prices/pairs/{pair}/latest",
            headers=self.initial_headers,
            params=kwargs,
        )

    def base_wap_latest(self, base: str, **kwargs) -> Dict:
        """Retrieves the latest VWAP & TWAP price for the specified base.

        :param base: The pair's base
        :key quote: The currency of the pair. Example: if pair is "eth_usd",
            then quote is "usd" (str)
        :key timeFormat: Time format to use for the timestamps
            (Options: milliseconds, ms, iso, iso8611) (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/prices/{base}/wap/latest",
            headers=self.initial_headers,
            params=kwargs,
        )

    def ticker_bid_ask_latest(self, pair: str, **kwargs) -> Dict:
        """Retrieves the latest market ticker Bid/Ask/Mid/Last for the
        specified pair.

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve market tickers.
            Example: exchange=bitfinex,bitstamp (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/tickers/{pair}/latest",
            headers=self.initial_headers,
            params=kwargs,
        )

    def ticker_bid_ask_historical(self, pair: str, **kwargs) -> Dict:
        """Retrieves the historical ticker, bid/ask/mid/last, for the specified
        pair.

        Note: This endpoint returns a max of 6 months of historical data. In order to get more
        than 6 months you must use the startDate & endDate parameters to move the time frame
        window to get the next n days/months of data.

        :param pair: The asset pair to look up
        :key exchange: The exchange(s) for which to retrieve market tickers.
            Example: exchange=bitfinex,bitstamp (str)
        :key startDate: Filter by ticker pairs after this date. (int)
        :key endDate: Filter by ticker pairs before this date. (int)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/tickers/{pair}/historical",
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_price_historical(self, address: str, **kwargs) -> Dict:
        """

        :param address: The token's smart contract address
        :key currency: The additional currency (other than ETH and USD) for which
            to return price info. (str)
        :key timeFormat: The time format to use for the timestamps: milliseconds/ms
            or iso/iso861. (str)
        :key timeInterval: The time interval to return the historical data in:
            by day (d),
            by hour (h), or
            by minute (m). (str)
        :key startDate: Filter by prices after this date. Note that the interval can not exceed
            6 months (d), 30 days (h) or 24 hours (m). (int)
        :key endDate: Filter by prices before this date. Note that the interval can not exceed
            6 months (d), 30 days (h) or 24 hours (m). (int)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/tokens/prices/{address}/historical",
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_price_latest(self, address: str, **kwargs) -> Dict:
        """Retrieves the latest price (and other market information) for the
        specified token.

        :param address: The token's smart contract address
        :key currency: The additional currency (other than ETH and USD)
            for which to return price info. (str)
        :return: The API response parsed into a dict
        """
        return self.raw_query(
            base_url=self.base_url,
            route=f"market/tokens/prices/{address}/latest",
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_rankings_historical(self, **kwargs) -> Dict:
        """Retrieves the top ranked tokens by a specific metric, with a
        lookback window.

        Useful for viewing token trends.

        :key direction: The sort order in which tokens are ranked
            (ascending or descending). (str)
        :key sortType: The metric used to rank the tokens (changeInPrice,
            currentPrice, marketCap, tokenVelocity, transactionVolume &
            uniqueAddresses). (str)
        :key topN: Number denominating top ranking tokens to return.
            Example: If given "5", results with return top 5 token
            rankings for past timeframe, where there are 5 results
            per day. (str)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self.raw_query(
            base_url=self.base_url,
            route="tokens/rankings/historical",
            headers=self.initial_headers,
            params=kwargs,
        )

    def token_rankings_latest(self, **kwargs) -> Dict:
        """Retrieves the top ranked tokens by a specific metric.

        :key direction: The sort order in which tokens are ranked
            (ascending or descending). (str)
        :key sortType: The metric used to rank the tokens (changeInPrice,
            currentPrice, marketCap, tokenVelocity, transactionVolume &
            uniqueAddresses). (str)
        :key timeInterval: The time interval to return the historical
            data in: by day (days) or by hour (hours). (str)
        :key type: The type(s) of tokens to include in the rankings
            (erc20, erc721, erc777, erc884, erc998) (str)
        :key page: The page number to return. (int)
        :key size: Number of records per page (int)
        :return: The API response parsed into a dict
        """

        self._check_chain_supported()
        return self.raw_query(
            base_url=self.base_url,
            route="tokens/rankings",
            headers=self.initial_headers,
            params=kwargs,
        )
